import numpy as np
import pickle
import glob
import re

def register_dots_coords(reg_data,hybridization,gene,all_raw_counts):
    
    """
    Function to register the coords of the dots according to the new coords
    of the corner of the images calculated after registration of the stitched 
    images.
    
    Parameters:
    -----------

    reg_data: dict 
        Dictionary containing the coords and the joining
        data calculated after the registration.
    hybridization: str 
        Name of the hybridization processed (Ex. Hybridization2)
    gene: str 
        Processed gene name (Ex. Gfap)
    all_raw_counts: dict 
        Dictionary with all the raw counting from all the hybridizations
    
    Returns:
    ---------

    all_raw_counts: dict 
        Dictionary with all the raw counting from all the hybridizations. 
        Added a subgroup containing the aligned coords.
    """
    
    
    tile_set = reg_data['micData'].tile_set.data
    tile_set = tile_set.ravel()
    for pos in all_raw_counts['selected_peaks_coords_not_aligned'][hybridization][gene].keys():  
        idx = np.where(tile_set==np.int(pos))[0][0]
        corner_coords = reg_data['joining']['corner_list'][idx][1]
        old_coords = all_raw_counts['selected_peaks_coords_not_aligned'][hybridization][gene][pos]
        if not np.all(old_coords==0):
            corrected_coords = old_coords + corner_coords
            all_raw_counts['selected_peaks_coords_aligned'][hybridization][gene][pos] = corrected_coords
    return all_raw_counts


def processing_combinations(hybridizations,aligned_peaks_dict):
    
    """
    Function used to create a list of tuples containing the pair hybridization
    and gene (Ex. (Hybridization1, Gfap)) that will be used to process the dots
    removal in parallel.
    
    Parameters:
    -----------

    hybridizations: list 
        List of the hybridization to process 
        (ex. [Hybridization1, Hybridization2]).
    aligned_peaks_dict: dict 
        Dictionary containing only the coords of the selected peaks after 
        registration (still with pos information).
                                
                                
    Returns:
    ---------
    combinations: list 
        List containing the pairs hybridization and gene 
        (Ex. (Hybridization1, Gfap))
    
    """  
    
    combinations_tmp=list()
    if isinstance(hybridizations,list):
        for hybridization in hybridizations:
            combinations_tmp.append([(hybridization,gene) for gene in aligned_peaks_dict[hybridization].keys()])
        combinations=[el for grp in combinations_tmp for el in grp]
    else:
        hybridization=hybridizations
        combinations=[(hybridization,gene) for gene in aligned_peaks_dict[hybridization].keys()]
    return combinations


def overlapping_images_grid(micData):
    """
    Function used to generate all possible overlapping images (neighborhood).
    
    Parameters:
    -----------

    micData: dict 
        Dictionary extracted from the reg_data and containing the infomation 
        regarding tiles organization
    
    Returns:
    ---------

    all_nhood: list 
        List with all the overlapping tiles combinations
    
    """
    
    tile_set=micData.tile_set.data
    Img_r,Img_c=tile_set.shape
    all_nhood=list()

    for r in np.arange(Img_r):
        for c in np.arange(Img_c):
            # Skip the position with the center tile ==-1 in the ImgSet
            if tile_set[r,c]!=-1:
                # I know the order of the neighboouring objects
                coordinates=[(r+i, c+j) for j in np.arange(-1,2) for i in np.arange(-1,2)
                             if ((0<=r+i<=Img_r-1) and (0<=c+j<=Img_c-1))]
                # Remove the pairs where -1 is in the ImgSet
                coordinates=[i for i in coordinates if tile_set[i]!=-1 ]
                if coordinates:
                    all_nhood.append(tuple(coordinates)) 
    return tuple(all_nhood)


def hood_creation(coords,pxl):
    
    """
    Function that creates the neighborood of coords that are considered 
    same RNA molecule. It is the region used to eliminate overlapping dots.
    
    Parameters:
    -----------

    coords: np.array 
        Array of the [r,c] of the dots.
    pxl: int
        Radius of pixel used to create the nhood.
    
    Return:
    ---------

    hood: set 
        Set with the coords of the neighborhood of the selected dot.
    
    """
    
    rg=np.arange(-(pxl),pxl+1)
    hood=tuple([(coords[0]+i,coords[1]+j) for i in rg for j in rg])
    return set(hood)


def dots_removing(overlapping_list,pxl):
    """
    Function used to identify matching dots that are present in overlapping 
    imaging areas.
    
    Parameters:
    -----------

    overlapping_list: list 
        List with the coords of the dots located in the overlapping region.                    
    pxl: int 
        Radius of pixel used to create the nhood.


    Returns:
    ---------

    removed_dots: set 
        Set with the coords of the removed dots.
    single_dots: set 
        Set with the coords of dots in the overlpping region that do not 
        overlap with any other dot.
    
    """
    
    removed_dots_coords=list()
    single_dots=set(overlapping_list.copy())
    remove_dot_step={}
    for coords in overlapping_list:
        if not {coords}.intersection(remove_dot_step):
            hood=hood_creation(coords,pxl)
            # Remove pxl to keep from hood
            hood=hood.difference({coords})
            single_dots=single_dots.difference(hood)
            remove_dot_step=set(overlapping_list).difference(set(single_dots))
    removed_dots=set(overlapping_list).difference(set(single_dots))
    return removed_dots,single_dots


def catch_duplicated_dots(combination, image_list,
                          overlapping_percentage,image_size,pxl):
    
    # dots_removing
    
    """
    Function used to identify dots that need to be removed or kept in the
    overlapping region between two images. The overlapping used is calculated
    according to the stitching information.
    
    Parameters:
    -----------

    combination: dict 
        Dictionary containing hybridization, gene, registation data of the 
        current hybridization and the registerd coords of the selected peaks 
        for the current gene.
    image_list: tuple 
        Tuple with the index of the the images with ooverlap
    overlapping_percentage: float 
        Percentage of overlap extracted from the Experimental_metadata.yaml.
    image_size: int 
        Size of the image (consider a square image).     
    pxl: int 
        Radius of pixel used to create the nhood.
    
    Returns:
    -----------

    overlapping_result_dict: dict 
        Dictionary with the coords of the dots that need to be removed or will 
        be kept for the selected pair of images.
    
    """
     
    # Extract the data
    micData = combination['reg_data_combination']['micData']
    joining = combination['reg_data_combination']['joining']
    aligned_peaks_dict_gene = combination['aligned_peaks_dict_gene']
    gene = combination['gene']
    hybridization = combination['hybridization']
    
    
    # Determine the image set from the stitched data
    img_set = micData.tile_set.data
    img_set_rav=img_set.ravel()

    # Create the overlapping results dict
    overlapping_result_dict=dict()
    overlapping_result_dict['row_dir']={}
    overlapping_result_dict['col_dir']={}

    overlapping_pixels=overlapping_percentage*image_size
    
    # Sort the list by row
    image_list=list(image_list)
    image_list.sort(key=lambda tup: tup[0]) 

    # work on rows
    rows=[el[0] for el in image_list]
    rows=set(rows)

    cols=[el[1] for el in image_list]
    cols=set(cols)

    overlapping=list()
    non_overlapping=list()
    counter=0

    # Process the images in rows
    for r in rows:
        row_grp=[ el for el in image_list if el[0] == r ]

        # Calculate the overlapping
        for p in np.arange(len(row_grp)-1):
            overlapping=list()
            non_overlapping=list()

            r,c=row_grp[p]
            pos_l=img_set[r,c]
            idx=np.where(img_set_rav==np.int(pos_l))[0][0]
            corner_left=joining['corner_list'][idx]
            corner_coords_left=corner_left[1]

            r,c=row_grp[p+1]
            pos_r=img_set[r,c]
            idx=np.where(img_set_rav==np.int(pos_r))[0][0]
            corner_right=joining['corner_list'][idx]
            corner_coords_right=corner_right[1]

            positions=[pos_l,pos_r]

            # Calculation expected coords in cols
            expected_corner_c=corner_coords_left[1]+image_size-overlapping_pixels
            c_error=np.abs(corner_coords_right[1]-expected_corner_c)
            r_error=np.abs(corner_coords_left[0]-corner_coords_right[0])

            # Remember I know the order of image acquisition
            # Determine the overlapping bounding-box
            r_TL=np.maximum(corner_coords_left[0],corner_coords_right[0])
            c_TL=corner_coords_right[1]
            TL=(r_TL,c_TL)

            r_BR=np.minimum(corner_coords_left[0]+image_size,corner_coords_right[0]+image_size)
            c_BR=corner_coords_left[1]+image_size
            BR=(r_BR,c_BR)


            # identify the dots in the overlapping region
            for pos in positions:
                if pos in aligned_peaks_dict_gene.keys():
                    dots=aligned_peaks_dict_gene[pos]

                    for dot in dots:
                            if (dot>=TL).all() and (dot<=BR).all():
                                overlapping.append(dot)
                            else:
                                non_overlapping.append(dot)

            # Fix the counting by removing the dots
            overlapping_list=[tuple(el) for el in overlapping]
            removed_dots,single_dots=dots_removing(overlapping_list,pxl)
            overlapping_result_dict['row_dir'][counter]={'tl':TL,'br':BR,'removed_dots':removed_dots,
                                                          'single_dots':single_dots}

            counter+=1


    counter=0 

    # Process the images in columns
    for c in cols:
        col_grp=[ el for el in image_list if el[1] == c ]

        # Calculate the overlapping
        for p in np.arange(len(col_grp)-1):
            overlapping=list()
            non_overlapping=list()

            r,c=col_grp[p]
            pos_t=img_set[r,c]

            idx=np.where(img_set_rav==np.int(pos_t))[0][0]
            corner_top=joining['corner_list'][idx]
            corner_coords_top=corner_top[1]
            r,c=col_grp[p+1]
            pos_b=img_set[r,c]

            idx=np.where(img_set_rav==np.int(pos_b))[0][0]
            corner_bottom=joining['corner_list'][idx]
            corner_coords_bottom=corner_bottom[1]

            positions=[pos_t,pos_b]

            # Calculation expecated coords in cols
            expected_corner_r=corner_coords_top[0]+image_size-overlapping_pixels
            r_error=np.abs(corner_coords_bottom[0]-expected_corner_r)
            c_error=np.abs(corner_coords_top[1]-corner_coords_bottom[1])

            # Remember I know the order of image acquisition
            # Determine the overlapping bounding-box
            r_TL=corner_coords_bottom[0]
            c_TL=np.maximum(corner_coords_bottom[1],corner_coords_top[1])
            TL=(r_TL,c_TL)


            r_BR=corner_coords_top[0]+image_size
            c_BR=np.minimum(corner_coords_bottom[1]+image_size,corner_coords_top[1]+image_size)
            BR=(r_BR,c_BR)


            # identify the dots in the overlapping region
            for pos in positions:
                if pos in aligned_peaks_dict_gene.keys():
                    dots=aligned_peaks_dict_gene[pos]

                    for dot in dots:
                            if (dot>=TL).all() and (dot<=BR).all():
                                overlapping.append(dot)
                            else:
                                non_overlapping.append(dot)

                # Fix the counting by removing the dots
                overlapping_list=[tuple(el) for el in overlapping]
                removed_dots,single_dots=dots_removing(overlapping_list,pxl)
                overlapping_result_dict['col_dir'][counter]={'tl':TL,'br':BR,'removed_dots':removed_dots,
                                                              'single_dots':single_dots}

                counter+=1

    
    return overlapping_result_dict


def all_dots_removal(combination,all_removed):

    """
    Function to remove the overlapping dots.
    
    Parameters:
    -----------

    combination: dict  
        Dictionary containing hybridization, gene, registation data of the 
        current hybridization and the registerd coords of the selected peaks 
        for the current gene.
    all_removed: dict 
        Dictionary with the coords of the dots that need to be removed or will 
        be kept for all the images

    Returns:
    -----------

    all_coords_cleaned: set 
        Set containing the coords of the dots after removal of the overlapping 
        ones. No image pos info is kept.
    removed_coords: set 
        Set containing the coords of the dots removed. No image pos info is kept.
    all_coords: set 
        Set containing the coords of all the aligned dots before removal
        of the overlapping ones. No image pos info is kept.
    
    
    """
    aligned_peaks_dict_gene = combination['aligned_peaks_dict_gene']
    gene = combination['gene']
    hybridization = combination['hybridization']
    
    # Create the set of the coords that will be removed
    directions_list=['row_dir','col_dir']
    removed_coords=set()
    for direction in directions_list:
        for idx in all_removed[hybridization][gene].keys():
                for pos in all_removed[hybridization][gene][idx][direction].keys():
                    if all_removed[hybridization][gene][idx][direction][pos]['removed_dots']:
                        removed_coords=removed_coords.union(all_removed[hybridization][gene][idx][direction][pos]['removed_dots'])

    # Create a set with the coords of all the dots
    all_coords=list()
    for pos in aligned_peaks_dict_gene.keys():
        all_coords.append([tuple(coords) for coords in aligned_peaks_dict_gene[pos]])

    all_coords=[el for coords in all_coords for el in coords]
    all_coords=set(all_coords)


    # Remove the coords from the dataset
    all_coords_cleaned=all_coords.difference(removed_coords)
    
    return all_coords_cleaned,removed_coords,all_coords

def function_to_run_dots_removal_parallel(combination,overlapping_percentage,image_size,pxl):
    
    """
    Function to run the overlapping dots removal in parallel
    
    Parameters:
    -----------

    combination: dict  
        Dictionary containing hybridization, gene, registation data of the 
        current hybridization and the registerd coords of the selected peaks 
        for the current gene.
    overlapping_percentage: float 
        Percentage of overlap extracted from the Experimental_metadata.yaml
    image_size: int 
        Size of the image (consider a square image)
    plx: int 
        Radius of pixel used to create the nhood.
    
    
    Returns:
    ---------

    all_coords_cleaned_dict: dict 
        Dictionary with the dots coords of the processed gene before and
        after removal plus the coords of the removed dots. 
    
    """
    # Determine hyb and gene processed
    gene = combination['gene']
    hybridization = combination['hybridization']
    
    micData = combination['reg_data_combination']['micData']
    
    # create the dictionaries
    all_removed=dict()
    all_coords_cleaned_dict=dict()
    all_removed[hybridization]={}
    all_coords_cleaned_dict[hybridization]={}

    all_removed[hybridization][gene]={}
    
    # Calculate all the neighborhood where to remove the dots
    all_nhood = overlapping_images_grid(micData)
    
    
    for idn,image_list in enumerate(all_nhood):
        overlapping_result_dict=catch_duplicated_dots(combination, image_list,
                          overlapping_percentage,image_size,pxl)
        all_removed[hybridization][gene][idn]=overlapping_result_dict
    
    all_coords_cleaned,removed_coords,all_coords = all_dots_removal(combination,all_removed)
    all_coords_cleaned_dict[hybridization][gene]={'all_coords_cleaned':all_coords_cleaned,
                                        'removed_coords':removed_coords,
                                        'all_coords':all_coords}
    
    return all_coords_cleaned_dict

def combine_raw_counting_results(flt_rawcnt_config,hybridizations_infos,
                                 experiment_infos,processing_experiment_directory,
                                stitched_reference_files_dir,reference_gene,add_slash):

    # Collect all the counts in a dictionary maintaining the pos references

    # Prepare the dictionary where to store the raw counts
    all_raw_counts = dict()
    all_raw_counts['selected_peaks_coords_aligned'] = dict()
    all_raw_counts['selected_peaks_coords_not_aligned'] = dict()
    all_raw_counts['remaining_data'] = dict()
    all_raw_counts['registration_data'] = dict()
    skip_genes_counting=flt_rawcnt_config['skip_genes_counting']
    skip_tags_counting=flt_rawcnt_config['skip_tags_counting']

    for hybridization in hybridizations_infos.keys():

        all_raw_counts['selected_peaks_coords_aligned'][hybridization] = dict()
        all_raw_counts['selected_peaks_coords_not_aligned'][hybridization] = dict()
        all_raw_counts['remaining_data'][hybridization]= dict()

        hyb_short = re.sub('Hybridization','hyb',hybridization)
        processing_hyb = experiment_infos['ExperimentName']+'_'+hyb_short
        hyb_dir = processing_experiment_directory+processing_hyb+add_slash

        # Determine the genes to stitch in the processing hybridization
        genes_processing = hybridizations_infos[hybridization].keys()

        # Determine the counting directories
        counting_gene_dirs_path = hyb_dir+flt_rawcnt_config['analysis_name']+'_'+processing_hyb +'_counting'+add_slash
        counting_gene_dirs = glob.glob(counting_gene_dirs_path+'*')

        # Load the _reg_data.pkl files with the information about the registration
        reg_fname = stitched_reference_files_dir + flt_rawcnt_config['analysis_name'] +'_'+ \
                                            processing_hyb+'_'+reference_gene+'_stitching_data_reg.pkl'
        reg_data = pickle.load(open(reg_fname,'rb'))


        all_raw_counts['registration_data'][hybridization]= reg_data

        for gene in genes_processing:
            if gene not in skip_genes_counting or [tag for tag in skip_tags_counting if tag in gene]:
                all_raw_counts['selected_peaks_coords_aligned'][hybridization][gene] = dict()
                all_raw_counts['selected_peaks_coords_not_aligned'][hybridization][gene] = dict()
                all_raw_counts['remaining_data'][hybridization][gene]= dict()


                # Determine the counting directory to process
                counting_files_dir = [pkl_dir for pkl_dir in counting_gene_dirs if gene in pkl_dir][0]
                counting_files_dir = counting_files_dir +add_slash

                # Get list of the files with counts
                counting_files = glob.glob(counting_files_dir+'*.pkl')


                # Load the raw counting data and combine them into a dictionary
                for counting_file in counting_files:
                    pos = int(counting_file.split('_')[-1].split('.')[0])
                    data = pickle.load(open(counting_file,'rb'))

                    all_raw_counts['selected_peaks_coords_not_aligned'][hybridization][gene][pos] = data['selected_peaks']
                    all_raw_counts['remaining_data'][hybridization][gene][pos] = dict()
                    for key in data.keys():
                        if key != 'selected_peaks':
                            all_raw_counts['remaining_data'][hybridization][gene][pos][key] = data[key]

                # Recalculate the coords but keep position reference
                all_raw_counts = register_dots_coords(reg_data,hybridization,gene,all_raw_counts)


    # Save all the data
    counting_data_name = processing_experiment_directory +experiment_infos['ExperimentName']+'_all_raw_counting_data.pkl'
    pickle.dump(all_raw_counts,open(counting_data_name,'wb'))
    
    return all_raw_counts



import numpy as np
import skimage.util as smutil
import skimage.exposure as smex
from skimage import img_as_float, img_as_uint
import glob


# import matplotlib.pyplot as plt
# from matplotlib.patches import Rectangle
# plt_available = True

plt_available = False
import logging
import warnings

from . import inout
from . import pairwisehelper as ph


"""Calculates the global corners, given the translations
for each tile. Also has functions to blend and stitch individual
images together in one large image, using the global corners.
"""

logger = logging.getLogger(__name__)


#################Functions to calculate global corners##############
def calc_corners_coord(tiles, transforms, md, nr_pixels, z_count):
    """Calculate global corner for each tile using coordinates.

    Calculates global corner for each tile using transforms and the
    microscope coordinates.

    Parameters:
    -----------

    tiles: list
        List of hdf5-references, references
        to the images in tile_file, with a
        reference for each tile. Here only used to check
        the length. If a tile does not have an
        associated image, its reference is None.
    transforms: np.array
        2d np array of size "number of tiles" by 2
        representing the y and x transform for each tile.
    md: object
        MicroscopeData object containing the tile set
        (to know the positioning of the tiles) and the y and
        x coordinates for each tile as documented by the
        microscope.
    nr_pixels: int
        Height and length of the tile in pixels, tile is assumed to be square.
    z_count: int
        The number of layers in one tile (size of the z
        axis). Should be 1 or None if the tile is 2D.

    Returns:
    --------
    : dict
        Dictionary containing keys corner_list and
        final_image_shape. Corner_list is a list of list, each
        list is a pair of an image number (int) and it's
        coordinates (numpy array containing floats).
        Final_image_shape is a tuple of size 2 or 3 depending on
        the number of dimensions and contains ints.
    """
    #Initialize the logger and arrays:
    logger.info("Filling corner list.")
    max_final_img       = np.zeros((2), dtype = float)
    min_final_img       = np.zeros((2), dtype = float)
    tile_inds           = range(len(tiles))
    temp_corner_list    = []

    #Get the masked array, used for checking for missing tiles
    masked_array    = np.ma.getmaskarray(md.tile_set.flat[:])
    logger.debug("masked_array: {}".format(masked_array))

    #Loop over the tiles in square order:
    for tile_ind in tile_inds:
        #Find out which image number belongs to the current tiles
        ind_coord       = md.tile_set.flat[:][tile_ind]
        #Check if tile is missing
        tile_missing    = masked_array[tile_ind]
        if not(tile_missing):
            logger.info("Calculating corner for tile index: {}  with "
                        + "tile number: {}"
                        .format(ind_coord, tile_ind))
            #Get the coordinates of the corner and adjust according
            #to the transform in transforms
            cur_corner = np.rint(np.array([md.y_coords[ind_coord],
                                           md.x_coords[ind_coord]])
                                 - transforms[tile_ind, -2:])
            logger.debug("Current transform, full: {} truncated: {}"
                         .format(transforms[tile_ind],
                                 transforms[tile_ind, -2:]))

            #Adjust the maximum size of the final image if necessary
            max_final_img   = np.maximum(max_final_img, cur_corner)
            min_final_img   = np.minimum(min_final_img, cur_corner)
            #Append the calculated corner to the temporary corner
            #list
            temp_corner_list.append(cur_corner)
        else:
            #If the tile we want to place is missing, the corner
            #will be flagged with np.nan values and added to the list
            logger.info("Tile is missing for tile index: {}  with "
                        + "tile number: {}"
                        .format(ind_coord, tile_ind))
            cur_corner = [np.nan, np.nan]
            temp_corner_list.append(cur_corner)
    #Adjust the temporary corner list according to the minimum of
    #the final image.
    #This way we get a new origin at min_final_image.
    logger.debug("min_final_img: {}".format(min_final_img))
    temp_corner_list   -= min_final_img
    #Place the temporary corner list in the global variable
    corner_list        = [[tile_inds[i], temp_corner_list[i]]
                                    for i in range(len(tile_inds))]
    #Adjust the shape of the final image
    if (z_count is None) or z_count == 1:
        final_image_shape  = tuple((max_final_img - min_final_img
                                    + nr_pixels).astype(int))
    else:
        final_image_shape  = tuple(np.append([z_count],
                                max_final_img - min_final_img
                                + nr_pixels).astype(int))
    #Give some feedback
    logger.info("Corners calculated")
    logger.debug("Corner_list: {}".format(corner_list))
    logger.info("Final image shape: {}".format(final_image_shape))

    return {'corner_list': corner_list,
            'final_image_shape': final_image_shape}


def apply_IJ_corners(tiles, corners, micData, nr_pixels):
    """Use corners from ImageJ for each tile to detemine corners.

    Parameters:
    -----------

    tiles: list
        List of hdf5-references, references
        to the images in tile_file, with a
        reference for each tile. Here only used to check
        the length. If a tile does not have an
        associated image, its reference is None.
    corners: list
        List of list, each list is a pair of an image number
        (int) and it's coordinates (numpy array containing
        floats).
    micData: object
        MicData object. Used to make an image list: the
        numbers of the images to be stitches sorted
        according to the tile indexes.
    nr_pixels: int
        Denoting size of the tile.

    Returns:
    --------
    : dict
        Same as calc_corners_coord() returns.
        Dictionary containing keys corner_list and
        final_image_shape. Corner_list is a list of list, each
        list is a pair of an image number (int) and it's
        coordinates (numpy array containing floats).
        Final_image_shape is a tuple of size 2 or 3 depending on
        the numer of dimensions and contains ints.
    """
    #Initialize the logger and arrays:
    logger.info("Filling corner list, using ImageJ")
    corner_list         = []
    max_final_img       = np.zeros((2))
    min_final_img       = np.zeros((2))
    tile_inds           = range(len(tiles))
    # Make a list of image numbers, matching with the numbers in the
    # image files
    flat_tile_set = micData.tile_set.flat[:]
    image_list = [micData.tile_nr[ind] if ind >= 0
                                    else -1 for ind in flat_tile_set]
    image_list = np.ma.masked_equal(image_list, -1)
    #Loop over the tiles in square order:
    for tile_ind in tile_inds:
        #Find out which image number belongs to the current tiles
        image_nr        = image_list[tile_ind]
        #Find the corner that matches the image number
        cur_corner = next(item[1] for item in corners
                            if item[0] == image_nr)
        logger.info("Index found (coord, image_nr, tile): {}, {}, {}"
                        .format(cur_corner, image_nr, tile_ind))
        #Adjust the maximum size of the final image if necessary
        max_final_img   = np.maximum(max_final_img, cur_corner)
        min_final_img   = np.minimum(min_final_img, cur_corner)
        corner_list.append([tile_ind, cur_corner])

    #TODO: Adjust corner to start at (0,0)
    logger.debug("min_final_img: {}".format(min_final_img))
    final_image_shape  = tuple((max_final_img - min_final_img
                                    + nr_pixels).astype(int))
    logger.info("Corners determined")
    logger.debug("Corner_list: {}".format(corner_list))
    logger.info("Final image shape: {}".format(final_image_shape))

    return {'corner_list': corner_list,
        'final_image_shape': final_image_shape}


########### Functions to join the tiles into to final image ############
def make_mask(joining, nr_pixels, blending_mask):
    """Calculate the mask that indicates where tiles overlap.

    This mask will have the same size a the final image.
    This function assigns a float (1.0, 2.0, 3.0 or 4.0) to each pixel,
    indicating if 1, 2, 3 or 4 tiles are going to overlap in this pixel.

    Parameters:
    -----------

    joining: dict
        Dictionary containing the corner list
        (with key: 'corner_list') with the tile indexes
        and their corresponding corners
    nr_pixels: int      
        Indicates the size of the tiles
    blending_mask: pointer
        Dataset in an hdf5 file containing a 2D numpy
        array. Array has the size of final image.
    """
    logger.info("Making blending mask")
    for i, corner in joining['corner_list']:
        if not(np.isnan(corner[0])):
            #Get the right part of mask
            cur_mask = blending_mask[int(corner[0]):int(corner[0]) + int(nr_pixels),
                            int(corner[1]):int(corner[1]) + int(nr_pixels)]
            # Place it back, after plus one
            blending_mask[int(corner[0]):int(corner[0]) + int(nr_pixels),
                            int(corner[1]):int(corner[1]) + int(nr_pixels)] = \
                             cur_mask + np.ones(cur_mask.shape, dtype=np.float64)
            #print(i,corner)
            #print('intervals:',int(corner[0]),int(corner[0]) + #int(nr_pixels),int(corner[1]),int(corner[1]) + int(nr_pixels))


def generate_blended_tile(temp_file, im_file, tiles, tile_file, corner_ind_coord, nr_pixels, tile_set,
                blend, linear_blending, ubyte,
                nr_dim = 2):
    """Blend the tile if necessary and then save it temp_file.

    Parameters:
    -----------

    temp_file: pointer
        Pointer to hdf5 file withv following groups:
        tiles, temp_masks, ubytes.
        Each group contains as many datasets as there are
        tiles, the datasets are named after the the tile
        index found in the first element of corner.
        This function places a blended tile and a corner in
        data set that matches the tile ind argument.
    im_file: pointer
            Pointer to hdf5 file with dataset "blending_mask"
            which contains a numpy array. blending_mask should
            be 1 where ther is no overlap and 2, 3 or 4 where
            the respective number of tiles overlap.
            Other datasets in this file are: final_image
            and temp_mask
    tiles: list
        List of hdf5-references, references
        to the images in tile_file, with a
        reference for each tile. If a tile does not have
        an associated image, its reference is None.
    tile_file: pointer
        hdf5 file object. The opened file containing the tiles to stitch.
    corner_ind_coord: list
        Contains two elements, the first one is an
        int representing the tile index, the second one is
        a numpy array containing the corner's coordinates.
    nr_pixels: int
        Denoting size of the tile.
    tile_set: np.array
        Masked numpy array. The shape of the array
        indicates the shape of the tile set.
    blend: bool
        When True blending will be applied,
        when false no blending at all will be applied.
    linear_blending: bool
        When True blending will be linear
        and when False, blending will be non-linear.
    ubyte: bool
        Ubyte image will be saved when True. Only full resolution image will be saved when False.
    nr_dim: int
        If 3, the code will assume three dimensional
        data for the tile, where z is the first dimension
        and y and x the second and third. For any other
        value 2-dimensional data is assumed. (default: 2)
    """
    # Get corner
    i = corner_ind_coord[0]
    corner = corner_ind_coord[1]
    # Load tile
    if nr_dim == 3:
        cur_tile = inout.load_tile_3D(tiles[i], tile_file)
    else:
        cur_tile = inout.load_tile(tiles[i], tile_file)
    # Check if tile is empty
    if cur_tile.size:
        logger.info("\nBlending tile: {}".format(i))
        logger.info("Corner position in final image: {}"
                    .format(corner))
        #Pick the right region of the image with corner
        ymin = int(corner[0])
        ymax = int(corner[0]) + nr_pixels
        xmin = int(corner[1])
        xmax = int(corner[1]) + nr_pixels

        logger.debug(" Tile {}, Cur region: {} {} {} {}"
                            .format(i, ymin, ymax, xmin, xmax))
        if blend:
            #Get current mask and region to paste the tile into
            cur_mask   = im_file['blending_mask'][ymin:ymax, xmin:xmax]

            #Blend
            blended_tile, cur_temp_mask  = perform_blending_par_proof(i,
                    cur_mask, cur_tile,
                    linear_blending, tiles, tile_set, nr_pixels)

            #Save the blended tile and the adjusted mask (mask is just saved for debugging)
            logger.debug('shape tile in temp file: {}'
                        .format(temp_file['blended_tiles'][str(i)].shape))
            logger.debug('shape current tile: {}'
                        .format(blended_tile.shape))
            temp_file['blended_tiles'][str(i)][:] = blended_tile
            temp_file['temp_masks'][str(i)][:]  = cur_temp_mask

            #Place the low resolution region back,
            if ubyte:
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore',
                                    category = UserWarning)
                    temp_file['ubytes'][str(i)][:] = \
                                smutil.img_as_ubyte(blended_tile)
            logger.info("Blended tile {} placed into temp_group."
                        .format(i))
        else:
            # Place the region back without blending
            temp_file['blended_tiles'][str(i)][:] = cur_tile

            #Place the low resolution region back
            if ubyte:
                temp_file['ubytes'][str(i)][:] = \
                                smutil.img_as_ubyte(cur_tile)
            logger.info("Tile {} placed into temp_group without "
                        + "blending.".format(i))
    else:
        logger.info("Skipped empty tile, tile: {}".format(i))


def non_linear_blending(x):
    """Define sigmoid for non-linear blending

    The steepness and half value of the curve are hardcoded here and
    good for a 10% overlap. For other overlaps flexibity in the
    steepness may be good to implement. (Halfpoint should be the
    same)

    Parameters:
    -----------

    x: np.array
        1d np-array, used as x values in the sigmoid curve

    Returns:
    --------

    y : np.array
        1d np-array, y values corresponding to x after applying
        sigmoid function on them
    """
    y = 1 / (1 + np.exp((-20 * (x - 0.5))))

    #logger.debug("Non linear blending x {}".format(x))
    #logger.debug("Non linear blending y {}".format(y))
    #plt.figure("alpha")
    #plt.plot(x,y,'-*')
    #plt.show()
    return y


def non_linear_blending_corner(x,y):
    """Calculate blending weights for pixels where four tiles overlap

    Parameters:
    -----------

    x: np.array
        1d numpy array, x and y should be of the same length. The
        distance from the corner in the x direction for each pixel.
    y: np.array
        1d numpy array, x and y should be of the same length. The
        distance from the corner in the x direction for each pixel.

    Returns:
    --------
    : np.array
        1d numpy array, same size as x and y. Weight for each pixel.
    """
    return x * y


def perform_blending_par_proof(tile_ind, cur_mask,
                        cur_tile, linear_blending,
                        tiles, tile_set, nr_pixels):
    """Blend a tile with the background around it.

    Each pixel in the overlap gets a weight depending on its distance
    from the border.

    Parameters:
    -----------

    tile_ind: int
        Index of the current tile
    cur_mask: np.array
        The mask denoting were in the
        picture tile have already been placed. A value of
        0 means no picture, a value of 1 means that a
        picture has been placed, a value of 2 means that
        there is overlap with the current tile.
    cur_tile: np.array
        The warped image to be placed..
    linear_blending: bool
        If true perform linear blending,
        otherwise non-linear blending.
    tiles: list
        List of hdf5-references, references
        to the images in tile_file, with a
        reference for each tile. Here only used to check
        if a tile exists, if a tile does not have an
        associated image, its reference is None.
    tile_set: np.array
        np-array representing the shape of the tile set
    nr_pixels: int
        Denoting size of the tile.

    Returns:
    --------

    cur_tile: np.array
        The blended cur_tile, same type and
        size as the input argument cur_type.
    cur_temp_mask: np.array
        Array like the cur_mask as it was
        passed to the function, but with the pixels that
        overlap replaced by their respective weights. This
        mask can be plotted for debugging.
    """

    logger.info("Performing blending on tile {}".format(tile_ind))
    # Find overlapping indexes in the blending mask.
    overlapping_inds = np.where(cur_mask > 1.0)
    logger.debug("overlapping inds: {}".format(overlapping_inds))
    logger.debug("size of cur_mask: {}".format(cur_mask.shape))

    if overlapping_inds[0].any() or overlapping_inds[1].any():

        cur_corner_tl   = np.array([0,0])
        cur_corner_br   = np.array([nr_pixels,nr_pixels])

        # Make lists for the 4 borders, with large numbers for borders that should not be used:
        # Order: [top, left, bottom, right]
        border_direct   = np.ones(4, dtype = int) * 9999
        border_indirect = np.ones(4, dtype = int) * 9999
        true_int = 0            # Value to use if overlap at a border exists
        unr_tile_ind = np.unravel_index(tile_ind, tile_set.shape)
        logger.debug("Checking neighbours of current tile: {}".format(unr_tile_ind))
        # Go through the 4 directly neighbouring tiles
        try:
            upper_neighbour = (unr_tile_ind[0] - 1, unr_tile_ind[1])
            upper_neighbour_ind = np.ravel_multi_index(upper_neighbour, tile_set.shape)
            if tiles[upper_neighbour_ind] is not None:
                # There is an upper neighbour
                border_direct[0] = true_int
                logger.debug("Direct neighbour top")
        except ValueError:
            # There is no upper neighbour
            pass
        try:
            left_neighbour_ind = (unr_tile_ind[0], unr_tile_ind[1] - 1)
            left_neighbour_ind = np.ravel_multi_index(left_neighbour_ind, tile_set.shape)
            logger.debug('neighbouring tile on left: {}'.format(tiles[left_neighbour_ind]))
            if tiles[left_neighbour_ind] is not None:
                # There is a left neighbour
                border_direct[1] = true_int
        except ValueError:
            pass
        try:
            lower_neighbour_ind = (unr_tile_ind[0] + 1, unr_tile_ind[1])
            lower_neighbour_ind = np.ravel_multi_index(lower_neighbour_ind, tile_set.shape)
            if tiles[lower_neighbour_ind] is not None:
                # There is a lower neighbour
                border_direct[2] = true_int
        except ValueError:
            pass
        try:
            right_neighbour_ind = (unr_tile_ind[0], unr_tile_ind[1] + 1)
            right_neighbour_ind = np.ravel_multi_index(right_neighbour_ind, tile_set.shape)
            if tiles[right_neighbour_ind] is not None:
                # There is a right neighbour
                border_direct[3] = true_int
        except ValueError:
            pass

        # Extra cases for indirect neighbours:
        try:
            indirect_neighbour_ind = (unr_tile_ind[0] - 1, unr_tile_ind[1] - 1)
            indirect_neighbour_ind = np.ravel_multi_index(indirect_neighbour_ind, tile_set.shape)
        except ValueError:
            pass
        else:

            if tiles[indirect_neighbour_ind] is not None:
                # There is an indirect neigbour on the top left
                border_indirect[0] = true_int
                border_indirect[1] = true_int
        try:
            indirect_neighbour_ind = (unr_tile_ind[0] - 1, unr_tile_ind[1] + 1)
            indirect_neighbour_ind = np.ravel_multi_index(indirect_neighbour_ind, tile_set.shape)
        except ValueError:
            pass
        else:
            if tiles[indirect_neighbour_ind] is not None:
                # There is an indirect upper neigbour on the right
                border_indirect[0] = true_int
                border_indirect[3] = true_int
        try:
            indirect_neighbour_ind = (unr_tile_ind[0] + 1, unr_tile_ind[1] - 1)
            indirect_neighbour_ind = np.ravel_multi_index(indirect_neighbour_ind, tile_set.shape)
        except ValueError:
            pass
        else:
            if tiles[indirect_neighbour_ind] is not None:
                # There is an indirect neigbour on the bottom left
                border_indirect[2] = true_int
                border_indirect[1] = true_int
        try:
            indirect_neighbour_ind = (unr_tile_ind[0] + 1, unr_tile_ind[1] + 1)
            indirect_neighbour_ind = np.ravel_multi_index(indirect_neighbour_ind, tile_set.shape)
        except ValueError:
            pass
        else:
            if tiles[indirect_neighbour_ind] is not None:
                # There is an indirect neigbour on the bottom right
                border_indirect[2] = true_int
                border_indirect[3] = true_int
        # Take the indirect and direct borders together:
        border_all = np.minimum(border_direct, border_indirect)
        logger.debug("Border direct: {}".format(border_direct))
        logger.debug("Border indirect: {}".format(border_indirect))
        logger.debug("Border all: {}".format(border_all))

        # For each pixel in the overlapping pixels calculate the distance from
        # the border and determine a weight based on the distance

        # Determine how many tiles overlap in the overlapping pixels
        overlapping_values = cur_mask[overlapping_inds]
        # Determine border distance to the top-left border and the bottom-right border:
        # So that there will be an np array of shape: (nr of pixels, 4), where each row
        # contains the distance to the corners in this order: [top, left, bottom, right]
        overlapping_pxls = np.array(overlapping_inds).T
        border_dist_list_raw = np.hstack((abs(overlapping_pxls - cur_corner_tl),
                                      abs(overlapping_pxls - cur_corner_br)))
        logger.debug("border_dist_list: {}".format(border_dist_list_raw))
        logger.debug("border_dist_list shape: {}".format(border_dist_list_raw.shape))

        # Take out irrelevant borders, add a large number so thay will not be taken
        # into account in the later analysis
        border_dist_list_indirect = border_dist_list_raw + border_indirect
        border_dist_list_direct = border_dist_list_raw + border_direct
        border_dist_list = border_dist_list_raw + border_all

        # Center of the tile
        center = int(nr_pixels / 2)
        logger.debug("center: {}".format(center))
        # Find all the pixels that have the center dist to one of the borders:
        # Center pixels will include pixels that form a plus centered
        # on the center of the tile and are part of an overlapping bit, like so:
        #  _________
        # |    |    |
        # |__     __|
        # |         |
        # |____|____|
        #

        center_pixel_inds = np.where(border_dist_list_raw == center)[0]
        logger.debug("center_pixel_inds: {}".format(center_pixel_inds))
        if center_pixel_inds.any():
            # Find the normalization factor for each border:
            # Look at the border distance of the pixels in the center
            center_pixels_dist = border_dist_list_direct[center_pixel_inds, :]
            # Get only the center pixels distances that belong to each border,
            # prevents taking distances into account that are to far to belong to
            # the corresponding border.
            # Set the pixels that are farther away than the center to 1,
            # to prevent them from being included in further analysis.
            center_pixels_dist[center_pixels_dist >= center] = 1
            # Find the maximum distance to each border within the center pixels.
            # This will be used as the maximum distance to the border, within
            # the overlap that belongs to that border.
            # (yields array of length 4)
            max_dist_pixels = np.amax(center_pixels_dist, axis = 0)
            logger.debug("Max border dist pixels: {}".format(max_dist_pixels))
            # Normalize with the found maximum distance for each border:
            border_dist_list_direct_norm = np.asfarray(border_dist_list_direct) \
                                                    / np.asfarray(max_dist_pixels)
            logger.debug("Closest border for each pixel: {}"
                            .format(np.argmin(border_dist_list_direct_norm,  axis = 1)))
            # Get the normalized distance to the closest border for each pixel.
            # We now have a 1D array of length nr of pixels, where each entry
            # represents the normalized distance from the pixel to the border
            # it belongs to.
            border_dist_norm = np.nanmin(border_dist_list_direct_norm, axis = 1)
            logger.debug("border_dist_norm: {}".format(border_dist_norm))
            logger.debug("border_dist_norm max, min: {} {}"
                            .format(np.amax(border_dist_norm),
                                    np.amin(border_dist_norm)))

            # Plots to check which pixels belong to which border,
            # uses closest_border.
            #closest_border = np.argmin(border_dist_list_direct_norm, axis=1)
            #cur_temp_mask1 = np.copy(cur_mask)
            #cur_temp_mask1[:,:] = 4.0
            #cur_temp_mask1[overlapping_inds] = np.array(closest_border)
            #plt.figure("Assigned borders")
            #plt.imshow(cur_temp_mask1, interpolation = 'none')
            #plt.show(block = False)
            #cur_temp_mask2 = np.copy(cur_mask)
            #cur_temp_mask2[overlapping_inds] = np.array(border_dist_norm)
            #plt.figure("border distance")
            #plt.imshow(cur_temp_mask2, 'gray', interpolation = 'none')
            #plt.show()

            # Plot the mask and blended tile
            # cur_temp_mask = np.copy(cur_mask)
            # cur_temp_mask[overlapping_inds] = np.array(border_dist_norm)
            # plt.figure("Normalized distance linear")
            # plt.imshow(cur_temp_mask, 'gray', interpolation = 'none')
            # plt.show()

            # Now calculate weights for the blending with direct neighbours
            if linear_blending:
                weights = border_dist_norm
            else:
                weights = non_linear_blending(border_dist_norm)
            logger.debug("weights max, min: {} {}"
                            .format(np.amax(weights), np.amin(weights)))
            # Plot the mask and blended tile
            # cur_temp_mask = np.copy(cur_mask)
            # cur_temp_mask[overlapping_inds] = np.array(weights)
            # plt.figure("Blending without corner blending")
            # plt.imshow(cur_temp_mask, 'gray', interpolation = 'none')
            # plt.show()
            #~ plt.figure("Blending tile")
            #~ plt.imshow(cur_tile, 'gray', interpolation = 'none')
            #~ plt.show()

            # Change the weights in the corners:
            weights = check_corner_blending(weights, border_dist_list, max_dist_pixels,
                            overlapping_values,
                            border_direct, border_indirect, center,
                            linear_blending)
        else:
            # We have only a corner overlapping, no sides
            logger.debug('{} has only a lonely corner and nothing else'.format(tile_ind))
            # Set weights to one to give them to check_corner_blending
            weights = np.ones(len(overlapping_inds[0]))
            # Blend the corner, with max_value set to two.
            weights = check_corner_blending(weights, border_dist_list, None,
                            overlapping_values,
                            border_direct, border_indirect, center,
                            linear_blending, max_value = 2)
        
        # Converted to float because data where save in uint16
        cur_tile=img_as_float(cur_tile)
        # Apply the weights to the pictures
        if cur_tile.ndim == 3:
            for i in range(cur_tile.shape[0]):
                cur_tile[i][overlapping_inds] *= np.array(weights)
                #~ plt.figure("Blending tile")
                #~ plt.imshow(cur_tile[i], 'gray', interpolation = 'none')
                #~ plt.show()
        else:
            cur_tile[overlapping_inds] *= np.array(weights)

        # Fill the temporary mask for debugging
        cur_temp_mask = np.copy(cur_mask)
        cur_temp_mask[overlapping_inds] = np.array(weights)

        # Plot the mask and blended tile
        #~ if tile_ind in [32, 33, 34, 60, 61, 62, 88, 89, 90]:
            #~ plt.figure("Blending tile")
            #~ plt.imshow(cur_tile, 'gray', interpolation = 'none')
            #~ plt.show()
    else:
        # If there is no overlap just return the original tile and
        # the temporary mask for debugging
        cur_temp_mask = np.copy(cur_mask)

    return cur_tile, cur_temp_mask


def check_corner_blending(weights, border_dist_list, max_dist_pixels,
                          overlapping_values, border_direct,
                          border_indirect, center, linear_blending,
                          max_value = 4):
    """Function to improve blending of differently overlapping corners.

    In a corner 2, 3 or 4 tile can overlap. Depending on how many tiles
    overlap and in wich direction they overlap the weights for a tile
    are adjusted by this function.

    Parameters:
    -----------

    weights: np.array
        1D numpy array. A weight for each overlapping
        pixel, as determined by the function
        perform_blending_par_proof (Weights depend on
        direct borders, or 1 if there is only a lonely
        corner)
    border_dist_list: np.array
        2D numpy array of shape: (nr of overlapping
        pixels, 4). Array containing for each pixel the
        distance to each of the four borders in this
        order [top, left, bottom, right], for borders
        that do not overlap to another
        tile the distance has been set to 999 or
        greater.
    max_dist_pixels: np.array
        Numpy array of shape: (1, 4). The maximum
        distance to each of the four borders in this
        order [top, left, bottom, right], for borders
        that do not overlap to another tile the max
        distance will be to 9999 or greater.
    overlapping_values: np.array
        1D numpy array of floats. An float value
        indicating how many tiles are overlapping
        for each overlapping pixel. This array has the
        same size as weights.
    border_direct: np.array
        1 by 4 numpy array. Array indicating for each
        border if there is a direct overlap.
        If there is overlap the value is 0, otherwise
        the value is 9999.
    border_indirect: np.array
        1 by 4 numpy array. Array indicating for each
        border if there is a indirect overlap.
        If there is overlap the value is 0, otherwise
        the value is 9999.
    center: int
        The center of th tile, this is the
        same number in the x and y direction.
    linear_blending: bool
        If true perform linear blending,
        otherwise non-linear blending.
    max_value: int
        The maximum value expected in the
        overlap. Default is 4, this is the overall
        maximum in the blending mask. When max_value
        is 2, we assume that the tile only overlaps
        in one or multiple corners. Because when only a
        corner of a tile is overlapping and there is
        no other overlap the maximum is 2. (Default: 4)

    Returns:
    --------

    weights: np.array
        1D numpy array. The array 'weights' that was passed
        to as an argument, but now adjusted for the corners.
    """
    logger.info("Performing corner blending...")
    # Find the corners
    # Define the borders that belong to each corner
    corner_defs = [[0,1], [0,3], [2,1],[2,3]]
    logger.debug("Border_dist_list: {}".format(border_dist_list[:,:]))
    for corner in corner_defs:
        # Init lonely corner
        lonely_corner = False
        # Find the indexes of the corner pixels
        corner_inds1 = np.where(border_dist_list[:,corner[0]] <= center)
        corner_inds2 = np.where(border_dist_list[:,corner[1]] <= center)
        big_corner_inds  = np.intersect1d(corner_inds1[0], corner_inds2[0])
        # Print the corners indexes
        logger.debug("Corner {} inds1: {}".format(corner, corner_inds1))
        logger.debug("Corner {} inds2: {}".format(corner, corner_inds2))
        logger.debug("Corner {} inds_intersect: {}".format(corner, big_corner_inds))
        # Determine if there is a corner where only 2 tiles overlap
        # (if this is the case lonely_corner will be set to True)
        # If 3 or four tiles overlap lonely corner is set to False.
        if ((border_indirect[corner[0]] < 999) and (border_direct[corner[0]] > 1)
            and (border_indirect[corner[1]] < 999) and (border_direct[corner[1]] > 1)):
            # There are only indirect tiles overlapping with this corner
            pos_corner_inds = np.where(overlapping_values == 2.0)[0]
            lonely_corner = True
        elif max_value == 2:
            # If max_value of 2 was passed, assume that the corner has
            # only 2 tiles overlapping
            pos_corner_inds = np.where(overlapping_values == 2.0)[0]
            lonely_corner = True
        else:
            # Else assume that more tiles are overlapping
            pos_corner_inds = np.where(overlapping_values > 2.0)[0]
        corner_inds  = np.intersect1d(big_corner_inds, pos_corner_inds)
        logger.debug("Corner {} inds_intersect pos: {}".format(corner, corner_inds))

        #If we found any overlap in the corner:
        if corner_inds.any():
            if lonely_corner:
                # 2 tiles are overlapping, but only in the corner
                # Calculate the Manhattan distance to the corner of the
                # current tile for each pixel belonging to this corning.
                border_dist_corner =  np.asfarray(border_dist_list[corner_inds, corner[0]] \
                                    + border_dist_list[corner_inds, corner[1]])
                logger.debug('LC border_dist_corner {}'.format(border_dist_corner))
                # Find the maximum distance for normalization
                max_dist_pixels_corner = float(max(border_dist_corner))
                logger.debug('LC max_dist_pixels_corner {}'.format(max_dist_pixels_corner))
                # Normalize and apply non-linear function if necessary
                if linear_blending:
                    weights_corner = border_dist_corner / \
                                        max_dist_pixels_corner
                else:
                    weights_corner = non_linear_blending(
                                            border_dist_corner / \
                                            max_dist_pixels_corner)
                #Place back new weights, overwrite old:
                weights[corner_inds] = weights_corner
                logger.debug('LC weights {}'.format(weights))
                logger.debug('LC max weights {}'.format(max(weights)))
            else:
                if np.mean(overlapping_values[corner_inds]) >= 3.5:
                    # 4 tiles overlap.
                    # If the mean of this corner is larger than 3.5, assume
                    # that 4 tiles overlap here.

                    # Pick the indexes of the corner where all 4 tiles
                    # actually overlap
                    corner_inds4 = np.where(overlapping_values == 4)
                    corner_inds  = np.intersect1d(corner_inds, corner_inds4[0])
                    logger.debug("Corner {} inds_intersect  4: {}".format(corner, corner_inds))
                    #Find the manhattan distance to corner at these indexes
                    border_dist_corner = border_dist_list[corner_inds, corner[0]] \
                                         + border_dist_list[corner_inds, corner[1]]
                    max_dist_pixels_corner = float(max(border_dist_corner))
                    # Normalize and apply non-linear function if necessary
                    if linear_blending:
                        weights_corner = border_dist_corner / \
                                            max_dist_pixels_corner
                    else:
                        y = np.asfarray(border_dist_list[corner_inds, corner[0]]) / \
                                                float(max_dist_pixels[corner[0]])
                        x = np.asfarray(border_dist_list[corner_inds, corner[1]]) / \
                                                float(max_dist_pixels[corner[1]])
                        logger.debug('x: {}  x max: {}'.format(x, max(x)))
                        logger.debug('y: {} y max: {}'.format(y, max(y)))
                        weights_corner = non_linear_blending_corner(x, y)
                    # Place back new weights:
                    weights[corner_inds] = weights_corner
                elif np.mean(overlapping_values[corner_inds]) < 3.5:
                    # 3 corners overlap.
                    # If the mean of this corner is smaller than 3.5, assume
                    # that 4 tiles overlap here.
                    corner_inds3    = np.intersect1d(corner_inds, np.where(overlapping_values == 3)[0])
                    #corner_inds     = corner_inds3
                    logger.debug("Corner {} inds_intersect 3: {}".format(corner, corner_inds3))
                    # Find which of the two borders is not a direct border:
                    special_border = None
                    special_border_list = np.where(border_direct > 1)[0]
                    logger.debug('special_border_list {}'.format(special_border_list))
                    for border in corner:
                        if border in special_border_list:
                            logger.debug('special_border found {}'.format(border))
                            special_border = border
                    if special_border is not None:
                        # Special border means that this border does
                        # not overlap with the tile next to it,
                        # therefore there are no overlapping pixels
                        # assigned to this border.
                        # Take all pixels that belong to the special
                        # border into account with the corner:
                        # Determine the maximum distance from the
                        # borders within this corner
                        max_dist_pixels_corner = np.amax(border_dist_list[corner_inds3, :], axis = 0)
                        # Determine new corner pixels based on the maximum distance
                        extra_corner_inds1 = np.where(border_dist_list[:, special_border]
                                                    <= max_dist_pixels_corner[special_border])[0]
                        logger.debug('extra_corner_inds1 {}'.format(extra_corner_inds1))
                        logger.debug('big_corner_inds {}'.format(big_corner_inds))
                        #Replace corner inds with new corner indexes, including the extra corner indexes:
                        corner_inds = np.intersect1d(extra_corner_inds1, big_corner_inds)
                        logger.debug('corner_inds shape {}'.format(corner_inds.shape))
                        # Replace corner inds 3, with part of the corner that is 3:
                        corner_inds3 = np.where(overlapping_values[corner_inds] == 3)[0]
                        logger.debug('corner_inds3 shape {}'.format(corner_inds3.shape))

                        # Get the manhattan distance where 3 tiles overlap
                        border_dist_corner3 = np.asfarray(
                                border_dist_list[corner_inds[corner_inds3], corner[0]] \
                                + border_dist_list[corner_inds[corner_inds3], corner[1]])
                        logger.debug('border_dist_corner {}'.format(border_dist_corner3))
                        max_man_dist_corner = float(max(border_dist_corner3))
                        logger.debug('max_man_dist_corner {}'.format(max_man_dist_corner))
                        # Also get the manhattan distance for all corner pixels,
                        # also where only 2 tiles overlap
                        border_dist_corner = np.asfarray(
                                    border_dist_list[corner_inds, corner[0]] \
                                    + border_dist_list[corner_inds, corner[1]])
                        # Then normalize this using the maximum of the manhattan
                        # distance in the area where 3 tiles overlap.
                        # So the part where 3 tiles overlap will be normalized in
                        # a normal way, but the parts where only 2 tiles overlap will
                        # not be normalized properly.
                        norm_manh_dist = border_dist_corner / max_man_dist_corner
                        # Then, adjust the norm_manh_dist to deal with the parts where two tiles overlap.
                        # Get the minimum and maximum border distance where all 3 tiles overlap
                        # (direct distance to the closest border)
                        max_corner3_dist = np.amax(
                                            border_dist_list[corner_inds[corner_inds3],:], axis = 0)
                        min_corner3_dist = np.amin(
                                            border_dist_list[corner_inds[corner_inds3],:], axis = 0)
                        logger.debug('max_corner3_dist min_corner3_dist {} {}'
                                    .format(max_corner3_dist, min_corner3_dist))
                        # Then used this min and max to adjust the weights of pixels that are closer to a border
                        # to zero and weights of pixels that are farther away from a border to 1.
                        for i in corner:
                            norm_manh_dist[border_dist_list[corner_inds, i]
                                                    < min_corner3_dist[i]] = 0.0
                            logger.debug('number of pixels closer to corner than end overlap 3: {}'
                                         .format(len(norm_manh_dist[border_dist_list[corner_inds, i]
                                                    < min_corner3_dist[i]])))
                            norm_manh_dist[border_dist_list[corner_inds, i]
                                                    > max_corner3_dist[i]] = 1.0
                            logger.debug('number of pixels farther away from corner than start overlap 3: {}'
                                     .format(len(norm_manh_dist[border_dist_list[corner_inds, i]
                                                    > max_corner3_dist[i]])))
                        # Make the weights non-linear if necessary
                        if linear_blending:
                            corner_weights = norm_manh_dist
                        else:
                            corner_weights = non_linear_blending(norm_manh_dist)
                        # Multiply the original weights in the corner with the  newly calculated corner weights,
                        # this way in the corner we get a combination of weights decreasing towards the direct border
                        # and weights decreasing towards the corner of the tile.
                        weights[corner_inds] *= corner_weights

    return weights


def make_final_image(joining, temp_file, im_file, nr_pixels):
    """Puts blended tiles at the correct position in the final image.

    Takes blended tiles as found in temp_file and "pastes" them at the
    correct position as indicated by joining in the final image. The
    final image is kept in im_file.

    Parameters:
    -----------

    joining: dict
        Containing corners for tiles
    temp_file: pointer
        Pointer to hdf5 object with the following groups:
        tiles,  temp_masks, ubytes.
        Each group contains as many datasets as there are
        tiles, the datasets are named after the the tile
        index found in the first element of corner.
        This function places a blended tile and a corner in
        data set that matches the tile ind argument.
    im_file: pointer
        Pointer to hdf5 object with dataset
        "blending_mask"
        which contains a numpy array. blending_mask should
        be 1 where ther is no overlap and 2, 3 or 4 where
        the respective number of tiles overlap.
        Other datasets in this file are: final_image
        and temp_mask
    nr_pixels: int
        Size of the tile, tile is assumed to be a square.
    """
    logger.info("Looping over all tiles and pasting them in the final "
                + "image...")
    for i, corner in joining['corner_list']:
        if not (np.isnan(corner[0])):  # Check for empty tile
            logger.info('Placing tile: {} in corner: {}'
                        .format(i, corner))
            # Pick the right region of the image using the corner
            ymin = int(corner[0])
            ymax = int(corner[0]) + nr_pixels
            xmin = int(corner[1])
            xmax = int(corner[1]) + nr_pixels
            logger.debug('i: {} ymin: {} ymax: {} xmin: {} xmax: {}'.
                         format(i, ymin, ymax, xmin, xmax))
            # Place the tile:
            if len(joining['final_image_shape']) == 3:
                im_file['final_image'][:, ymin:ymax, xmin:xmax] += temp_file['blended_tiles'][str(i)]
                if 'final_image_ubyte' in im_file:
                    im_file['final_image_ubyte'][:, ymin:ymax, xmin:xmax] += temp_file['ubytes'][str(i)]
            else:
                im_file['final_image'][ymin:ymax, xmin:xmax] += temp_file['blended_tiles'][str(i)]
                if 'final_image_ubyte' in im_file:
                    im_file['final_image_ubyte'][ymin:ymax, xmin:xmax] += temp_file['ubytes'][str(i)]
            im_file['temp_mask'][ymin:ymax, xmin:xmax] += temp_file['temp_masks'][str(i)]

def paste_in_final_image_MPI(joining, temp_file, im_file, tile_ind, nr_pixels):
    """Puts one blended tile at the correct position in the final image.

    Takes the blended tile as found at tile_ind in temp_file and
    "pastes" them at the correct position as indicated by joining in
    the final image. The final image is kept in im_file.

    Parameters:
    -----------

    joining: dict
        Containing corners for tiles
    temp_file: pointer
        Pointer to hdf5 object with the following groups:
        tiles,  temp_masks, ubytes.
        Each group contains as many datasets as there are
        tiles, the datasets are named after the the tile
        index found in the first element of corner.
        This function places a blended tile and a corner in
        data set that matches the tile ind argument.
    im_file: pointer
        Pointer to hdf5 object with dataset
        "blending_mask"
        which contains a numpy array. blending_mask should
        be 1 where ther is no overlap and 2, 3 or 4 where
        the respective number of tiles overlap.
        Other datasets in this file are: final_image
        and temp_mask
    tile_ind: int
        Index of the tile that should be placed
    nr_pixels: int
        Size of the tile, tile is assumed to be a square.
    """
    # Find the right corner
    cur_corner = next(([i, corner] for i, corner in joining['corner_list']
                        if (i == tile_ind)), [0, [np.nan, np.nan]])[1]
    logger.info('Pasting tile {} in final image, in corner: {}'
                .format(tile_ind, cur_corner))
    logger.debug('tile_ind: {} cur_corner: {}'.format(tile_ind, cur_corner))
    if not (np.isnan(cur_corner[0])):            # Check for empty tile
        # Pick the right region of the image using the corner
        ymin = int(cur_corner[0])
        ymax = int(cur_corner[0]) + nr_pixels
        xmin = int(cur_corner[1])
        xmax = int(cur_corner[1]) + nr_pixels
        logger.debug('Global corners: {} {} {} {}'.
                     format(ymin, ymax, xmin, xmax))
        # Place the tile:
        if len(joining['final_image_shape']) == 3:
            im_file['final_image'][:, ymin:ymax, xmin:xmax] += temp_file['blended_tiles'][str(tile_ind)]
            if 'final_image_ubyte' in im_file:
                im_file['final_image_ubyte'][:, ymin:ymax, xmin:xmax] += temp_file['ubytes'][str(tile_ind)]
        else:
            im_file['final_image'][ymin:ymax, xmin:xmax] += temp_file['blended_tiles'][str(tile_ind)]
            if 'final_image_ubyte' in im_file:
                im_file['final_image_ubyte'][ymin:ymax, xmin:xmax] += temp_file['ubytes'][str(tile_ind)]
        if 'temp_mask' in im_file:
            im_file['temp_mask'][ymin:ymax, xmin:xmax] += temp_file['temp_masks'][str(tile_ind)]


def assess_overlap(joining, tiles, tile_file, contig_tuples):
    """Calculate the covariance of the overlap

    This function is usefull to see the quality of the final overlap.
    It always works with flattened (max projected) tiles.

    Parameters:
    -----------

    joining: dict      
        Dictionary containing the corner list
        (with key: 'corner_list') with the tile indexes
        and their corresponding corners
    tiles: list
        List of hdf5-references, references
        to the images in tile_file, with a
        reference for each tile.
        If a tile does not have an associated image,
        its name is None.
    tile_file: pointer
        hdf5 file object. The opened file containing the
        tiles to stitch.
    contig_tuples: list
        List of tuples. Each tuple is a tile pair.
        Tuples contain two tile indexes denoting these
        tiles are contingent to each other.

    Returns:
    --------

    xcov: float
        The cross covariance between the overlapping parts of the two images.
    """
    logger.info("Assessing overlap...")
    allow_plot = False
    xcov_list = []
    sorted_corners = sorted(joining['corner_list'])
    logger.debug('sorted corner list: {}'.format(sorted_corners))
    for pair in contig_tuples:
        xcov = np.nan
        ind1 = min(pair)
        ind2 = max(pair)
        if tiles[ind1] is not None and tiles[ind2] is not None:
            if abs(ind1 - ind2) == 1:
                #Overlap on left or right
                logger.debug(("Overlap on right of tile {0} and left of"
                                    + " tile {1}").format(ind1, ind2))
                corner1 = sorted_corners[ind1][1]
                corner2 = sorted_corners[ind2][1]
                #Calculate overlap indexes
                logger.debug("x values of, ind {}, x-coord: {}  ind {}, x-coord: {}"
                                .format(ind1,  corner1[1], ind2, corner2[1]))
                logger.debug("y values of, ind {}, x-coord: {}  ind {}, x-coord: {}"
                                .format(ind1,  corner1[0], ind2, corner2[0]))
                overlap_ind_x   = int(corner1[1] - corner2[1])
                overlap_ind_y   = int(corner1[0] - corner2[0])
                logger.debug("Overlap index, x: {} y: {}".
                                format(overlap_ind_x, overlap_ind_y))
                # Loading tiles in function call, loading the flattened version of the tiles.
                overlap1, overlap2 = ph.get_overlapping_region(inout.load_tile(tiles[ind1], tile_file),
                                        inout.load_tile(tiles[ind2], tile_file),
                                        overlap_ind_x, overlap_ind_y, 'left')
                #For subplotting the overlaps later:
                plot_order = np.ones((1,3))
            else:
                #Overlap on top or bottom
                logger.debug(("Overlap on bottom of tile {0} and top of"
                                    + "tile {1}").format(ind1, ind2))
                corner1 = sorted_corners[ind1][1]
                corner2 = sorted_corners[ind2][1]
                #Calculate overlap indexes
                logger.debug("x values of, ind {}, x-coord: {} cur_corner ind {}, x-coord: {}"
                                .format(ind1,  corner1[1], ind2, corner2[1]))
                logger.debug("y values of, ind {}, y-coord: {}  ind {}, y-coord: {}"
                                .format(ind1,  corner1[0], ind2, corner2[0]))
                overlap_ind_x   = int(corner1[1] - corner2[1])
                overlap_ind_y   = int(corner1[0] - corner2[0])
                logger.debug("Overlap index, y: {} x: {}".
                                format(overlap_ind_y, overlap_ind_x))
                overlap1, overlap2 = ph.get_overlapping_region(inout.load_tile(tiles[ind1], tile_file),
                                        inout.load_tile(tiles[ind2], tile_file),
                                        overlap_ind_x, overlap_ind_y, 'top')
                #For subplotting the overlaps later:
                plot_order = np.ones((3,1))


            if allow_plot:
                overlap_rgb         = np.zeros((overlap1.shape[0], overlap1.shape[1], 3))
                overlap_rgb[:,:,0]  = smex.rescale_intensity(overlap1)
                overlap_rgb[:,:,1]  = smex.rescale_intensity(overlap2)

                #overlap_masked          = np.copy(overlap_rgb)
                #overlap_masked[:,:,0][overlapping_inds]    *=  1.0 - np.array(weights)
                #overlap_masked[:,:,1][overlapping_inds]    *= np.array(weights)

            cor_coeff, mono_color = ph.xcov_nd(overlap1, overlap2)
            if not(mono_color):
                xcov = cor_coeff

            if mono_color:
                logger.debug("The overlapping images are mono colored, "
                             + "xcov: {}".format(cor_coeff))
                title_text = "Mono color, xcov: {}".format(cor_coeff)
            else:
                logger.debug("Normalized cross covariance of overlap: "
                             + "{}".format(cor_coeff))
                title_text = "XCov: {}".format(cor_coeff)

            if allow_plot:
                inout.display_tiles([overlap1,
                                overlap2,
                                overlap_rgb],
                                plot_order, fig_nr = "overlap check",
                                maximize = True, main_title = title_text,
                                rgb = True)
        xcov_list.append(xcov)
    return xcov_list


def generate_blended_tile_npy(corner_ind_coord,stitching_files_dir,
                              blended_tiles_directory, masked_tiles_directory,
                              analysis_name,processing_hyb,reference_gene,
                              micData, tiles, nr_pixels,
                linear_blending):

    
    """
    Blend the tile if necessary and then save it temp blended folder.
    Modification of the generate_blended_tile that run using .npy files
    and doesn't save the data in a hdf5 file

    Parameters:
    ------------
    
    corner_ind_coord: list
        Contains two elements, the first one is an
        int representing the tile index, the second one is
        a numpy array containing the corner's coordinates.
    stitching_files_dir: str
        Path to the files to stitch
    analysis_name: str 
        Name of the current analysis
    blended_tiles_directory: str
        Path to the directory where to save the blended tiles    
    masked_tiles_directory: str
        Path to the directory with the masks                                       
    processing_hyb: str
        Name of the hybridization processed
    reference_gene: str
        Name of the gene to be stitched
    blending_mask: np.array
        Array containing the blending mask used for blending the images     
    micData: object
        MicroscopeData object. Contains coordinates of
        the tile corners as taken from the microscope.                   
    tiles: np.array
        Array with tile number. -1 correspond to missing tile      
    nr_pixels: int
        Denoting size of the tile. 
    linear_blending: bool
        When True blending will be linear and when False, blending will be non-linear.
                    
    
    
    """
       
    stitching_files_list = glob.glob(stitching_files_dir+'*.npy')
    masked_files_list = glob.glob(masked_tiles_directory+'*.npy')
    
    # Get corner
    tile_ind = corner_ind_coord[0]
    corner = corner_ind_coord[1]
    
    # Get the reference address of the tile to load
    tile_ref = tiles[tile_ind]
    
    # Check if the tile is existing
    if tile_ref != -1:
        
        # Get the tile number
        tile_number = '_'+str(tile_ref)+'.npy'
        
        # Load the tile
        tile_path = [tile_p for tile_p in stitching_files_list if tile_number in tile_p][0]
        cur_tile = np.load(tile_path)
        
        # Load the mask
        mask_ref = '_'+str(tile_ind)+'.npy'
        mask_path = [mask_p for mask_p in stitching_files_list if mask_ref in mask_p]
        if mask_path:
            cur_mask = np.load(mask_path[0])
        
            #Blend
            blended_tile, cur_temp_mask  = perform_blending_par_proof(tile_ind,
                    cur_mask, cur_tile,
                    linear_blending, tiles, micData.tile_set, nr_pixels)
        
        else:
            blended_tile = cur_tile
        
        # Convert the blended tile to uint16 to be able to save it in the 
        # prefilled hdf5
        # if blended_tile.max() >1 :
        #     blended_tile = blended_tile/blended_tile.max()

        # blended_tile[blended_tile<-0] = 0        
        # blended_tile = img_as_uint(blended_tile)

        # Save the blended tiles
        fname = blended_tiles_directory + analysis_name+'_'+processing_hyb+'_'+reference_gene+'_blended_tile_pos'+tile_number
        np.save(fname,blended_tile)


# Write the blended image in the hdf5 file
def make_final_image_npy(joining, stitching_file, blended_tiles_directory, tiles,gene, nr_pixels):
    
    """
    Puts blended tiles at the correct position in the final image.
    Modified version of the make_final_image that works on .npy stored
    images. Works only in 2D.

    Takes blended tiles as found in the blended_tiles_directory and "pastes" 
    them at the correct position as indicated by joining in the final image
    in the hdf5 file. 

    Parameters:
    -----------

    joining: dict
        Containing corners for tiles
    stitching_file: pointer
        Pointer to hdf5 object with the following groups:
    blended_tiles_directory: str
        Path to the directory where to save the blended tiles                    
    tiles: np.array
        Array with tile number. -1 correspond to missing tile
    gene: str
        Name of the gene to be stitched
    nr_pixels: int
        Size of the tile, tile is assumed to be a square.
    
    """
    
    
    blended_files_list = glob.glob(blended_tiles_directory+'*.npy')
    
    for i, corner in joining['corner_list']:
         
            tile_ref = tiles[i]
            
            if tile_ref != -1:
            # Get the tile number
                
                tile_number = '_'+str(tile_ref)+'.npy'
            
                image_path = [ipath for ipath in blended_files_list if tile_number in ipath][0]

                img = np.load(image_path)

                # Pick the right region of the image using the corner
                ymin = int(corner[0])
                ymax = int(corner[0]) + nr_pixels
                xmin = int(corner[1])
                xmax = int(corner[1]) + nr_pixels

                # Place the tile:
                stitching_file[gene]['StitchedImage']['final_image'][ymin:ymax, xmin:xmax] += img          
                stitching_file.flush()


from setuptools import setup, find_packages

__version__ = "0.0.2"
exec(open('pysmFISH/_version.py').read())

setup(
    name="pysmFISH",
    version=__version__,
    packages=find_packages(),
    python_requires='>3.6',
    install_requires=[
        'numpy',
        'scipy',
        'scikit-image',
        'scikit-learn',
        'matplotlib',
        'nd2reader==2.1.3',
        'sympy',
        'loompy',
        'ruamel.yaml',
        'h5py',
        'dask',
        'distributed',
        'mpi4py'
    ],
    entry_points='''
        [console_scripts]
        add_coords_to_experimental_metadata=add_coords_to_experimental_metadata
        apply_stitching=apply_stitching
        dots_coords_correction=dots_coords_correction
        preprocessing_script=preprocessing_script
        process_standalone_experiment=process_standalone_experiment
        reference_registration=reference_registration
        Run_dots_counting=Run_dots_counting
        staining_segmentation=staining_segmentation
    ''',

    author="Simone Codeluppi",
    author_email="simone.codeluppi@gmail.com",
    keywords=["spatial transcriptomics", "singlecell", "bioinformatics", "transcriptomics"],
    description="Analysis of large spatial transcriptomics data (osmFISH).",
    license="MIT",
    url="https://linnarssonlab.org/osmFISH/",
)



import h5py
import numpy as np
from skimage import filters,io,img_as_float,exposure,morphology,segmentation,measure,feature,color
from scipy import ndimage as nd
import itertools
import pickle
import multiprocessing

from pysmFISH import utils
from pysmFISH import object_based_segmentation

def staining_segmentation():

    """
    This script will segment the selected staining and output the identified 
    objects.

    All the parameters are entered via argparse.

    Parameters:
    -----------

    scheduler: string
        tcp address of the dask.distributed scheduler (ex. tcp://192.168.0.4:7003). 
        default = False. If False the process will run on the local computer using nCPUs-1
    path: string
        Path to the processing directory
    processing_file: string
        Path to the hdf5 file with the staning to process
    segmentation_staining: string
        Staining to be segmented

    """

    # Inputs of the function
    parser = argparse.ArgumentParser(description='Segmentation script')
    parser.add_argument('-scheduler', default=False, help='dask scheduler address ex. tcp://192.168.0.4:7003')
    parser.add_argument('-path', help='processing directory')
    parser.add_argument('-processing_file', help='path to the file with the \
                        staning to process')
    parser.add_argument('-segmentation_staining', help='staining to be \
                        segmented')
    
    

    args = parser.parse_args()
    
    # Directory to process
    processing_directory = args.path

    # File to process
    processing_file = args.processing_file

    # staining to segment
    segmentation_staining = args.segmentation_staining

    # Dask scheduler address
    scheduler_address = args.scheduler
    
    if scheduler_address:
        # Start dask client on server or cluster
        client=Client(scheduler_address)

    else:
        # Start dask client on local machine. It will use all the availabe
        # cores -1

        # number of core to use
        ncores = multiprocessing.cpu_count()-1
        cluster = LocalCluster(n_workers=ncores)
        client=Client(cluster)

    # Determine the operating system running the code
    os_windows, add_slash = utils.determine_os()

    # Check training slash in the processing directory
    processing_directory=utils.check_trailing_slash(processing_directory,os_windows)


    segmentation_parameters = utils.general_yaml_parser(processing_directory+'Staining_segmentation.config.yaml')

    # Chunking parameters
    chunk_size = segmentation_parameters[segmentation_staining]['image_chunking_parameters']['chunk_size']
    percent_padding = segmentation_parameters[segmentation_staining]['image_chunking_parameters']['percent_padding']

    # Segmentation parameters
    trimming = segmentation_parameters[segmentation_staining]['segmentation_parameters']['trimming']
    min_object_size = segmentation_parameters[segmentation_staining]['segmentation_parameters']['min_object_size']
    disk_radium_rank_filer = segmentation_parameters[segmentation_staining]['segmentation_parameters']['disk_radium_rank_filer']
    min_distance = segmentation_parameters[segmentation_staining]['segmentation_parameters']['min_distance']   
    threshold_rel = segmentation_parameters[segmentation_staining]['segmentation_parameters']['threshold_rel']    

    # Load the image (will be modified after the change to hdf5 input)
    img = io.imread(processing_file)

    # Image chunking
    nr_chunks,nc_chunks,Coords_Chunks_list, Coords_Padded_Chunks_list,r_coords_tl_all_padded,\
                c_coords_tl_all_padded,r_coords_br_all_padded,c_coords_br_all_padded = \
                object_based_segmentation.image_chunking(img,chunk_size,percent_padding)
            
            
    # Create the chunks idx
    Chunks_idxs_linear=np.arange(len(Coords_Padded_Chunks_list),dtype='int32')

    # Distribute the chunks idx and distridute them in an array according to the position
    # in the chunked image
    Chunks_idxs=Chunks_idxs_linear.reshape(nr_chunks,nc_chunks)  

    # Flatten the array for make it easier the creation of the coords combination
    Chunks_idxs_rows=np.ravel(Chunks_idxs)
    Chunks_idxs_cols=np.ravel(Chunks_idxs,order='F')

    # Calculate coords of the overlapping chunks
    Overlapping_chunks_coords=list()
    counter=0
    left_pos=Chunks_idxs_rows[0]
    for el in Chunks_idxs_rows[1:]:
        if counter < nc_chunks-1:
            Coords_left=Coords_Padded_Chunks_list[left_pos]
            Coords_right=Coords_Padded_Chunks_list[el]
            row_tl=Coords_left[0]
            row_br=Coords_left[1]
            col_tl=Coords_right[2]
            col_br=Coords_left[3]
            Overlapping_chunks_coords.append((row_tl,row_br,col_tl,col_br))
            left_pos=el
            counter+=1
        else:
            left_pos=el
            counter=0
            
    counter=0
    top_pos=Chunks_idxs_cols[0]
    for el in Chunks_idxs_cols[1:]:
        if counter < nr_chunks-1:
            Coords_top=Coords_Padded_Chunks_list[top_pos]
            Coords_bottom=Coords_Padded_Chunks_list[el]
            
            row_tl=Coords_bottom[0]
            row_br=Coords_top[1]
            col_tl=Coords_top[2]
            col_br=Coords_top[3]
            Overlapping_chunks_coords.append((row_tl,row_br,col_tl,col_br))
        
            counter+=1
            top_pos=el
        else:
            top_pos=el
            counter=0

    # Now i use this approach for testing. If the image gets to big to fit in RAM
    # then save the files and load them separately in each node
    chunked_image_seq = list()
    for coords in Coords_Padded_Chunks_list:
        chunked_image_seq.append(img[coords[0]:coords[1],coords[2]:coords[3]])

    # Run the segmentation
    futures_processes = client.map(object_based_segmentation.polyT_segmentation,chunked_image_seq,
                                min_object_size=min_object_size,
                                min_distance=min_distance,
                                disk_radium_rank_filer=disk_radium_rank_filer,
                                threshold_rel=threshold_rel,
                                trimming=trimming)

    Objects_list = client.gather(futures_processes)

    # Recalculate labels and coords

    processed_images_data = dict()

    max_starting_label = 0
    total_data_dict = dict()

    for idx, objs_chunk in enumerate(Objects_list):
        for label ,cvalues in objs_chunk.items():
                new_label=max_starting_label+1
                coords = Coords_Padded_Chunks_list[idx][0::2]
                total_data_dict[new_label] = cvalues+coords
                max_starting_label = new_label


    # Calculate all the intersecting objects
    futures_processes = client.map(object_based_segmentation.OverlappingCouples,Overlapping_chunks_coords,
                                TotalDataDict = total_data_dict)
        

    All_intersecting = client.gather(futures_processes)


    # Put together the couple with the same label for multiple intersection
    # for the labels of objects where there is intersection between multiple regions
    # Then scatter all of them and calculate intersection

    # Combine the results from the parallel processing
    flatten_couple = [el for grp in All_intersecting for el in grp] 
    # Remove duplicates
    flatten_couple=list(set(flatten_couple))

    # Create a list of the labels (removing the repeats)
    singles=list()
    [singles.append(x) for cpl in flatten_couple for x in cpl]
    singles=list(set(singles))

    # Identify the couples containing singles
    Combined_all_singles=list()
    for item in singles:
        Combined_single=list()
        for couple in flatten_couple:
            if item in couple:
                Combined_single.append(couple)
        Combined_all_singles.append(Combined_single)

    if Combined_all_singles:
        # Combine all the intersecting labeles
        start=Combined_all_singles[0]
        ComparisonList=Combined_all_singles[1:].copy()
        #merged=start.copy()
        merged=list()
        SavedCombinations=list()
        tmp_list=ComparisonList.copy()
        KeepGoing=True
        Loop=0
        while KeepGoing:
            Loop+=1


            for idx,el in enumerate(ComparisonList):

                if set(start).intersection(set(el)):
                    #merged=list(set(merged)|set(el))
                    [merged.append(x) for x in el]
                    tmp_list = [e for e in tmp_list if e != el]


            intersection=list(set.intersection(set(merged),set(start)))  
            if intersection:
                merged=list(set.union(set(merged),set(start)))
                #merged=list(set(merged))
                start=merged.copy()
                merged=list()
                ComparisonList=tmp_list.copy()
                #tmp_list.append(merged)
            else:
                SavedCombinations.append(start)
                start=tmp_list[0]
                tmp_list=tmp_list[1:]
                ComparisonList=tmp_list.copy()


            if len(tmp_list)<1:
                [SavedCombinations.append(x) for x in tmp_list]
                KeepGoing =False
        
        # Remove all the duplicated labeled that intersect
        # in this case the labeled are merged. It will be nice to run an extra
        # segmentation on the merged objects
        # If it is too slow this step can be parallelised

        SavedLab_list=list()
        CleanedDict=total_data_dict.copy()
        for couple in SavedCombinations:
            SaveLab, RemoveLabs,NewCoords=object_based_segmentation.IntersectionCouples(couple,total_data_dict)
            SavedLab_list.append(SaveLab)
            for lab in RemoveLabs:
                del CleanedDict[lab]
            CleanedDict[SaveLab]=NewCoords
    else:
        CleanedDict=total_data_dict 


    # Calculate all objects properties
    all_objects_list = [(key,coords) for key,coords in CleanedDict.items()]

    futures_processes = client.map(object_based_segmentation.obj_properties_calculator,all_objects_list)

    all_objects_properties_list = client.gather(futures_processes)


    # convert the list to a dictionary
    all_objects_properties_dict = { k: v for d in all_objects_properties_list for k, v in d.items() }

    # Save all the objects
    segmented_objs_fname = processing_directory + 'segmented_' + segmentation_staining + '_all_objs_properties.pkl'
    pickle.dump(all_objects_properties_dict,open(segmented_objs_fname,'wb'))


if __name__ == "__main__":
    staining_segmentation()


import argparse

from pysmFISH import utils
from pysmFISH.stitching_package import hybregistration as hr

def reference_registration():

    """
    This script is used to register the stitched reference channels for the processed
    hybridization. The comparison is sequential (from Hyb1-->HybN) and not all the
    hybridization steps are required.

    The output are pickle files with the recalculated corners according to the 
    registration

    The input parameters are entered via argparse

    Parameters:
    -----------

    path: string. 
        Exact path to the folder with the stitched .sf.hdf5
    reference_gene: string. 
        Reference gene used for stitching
    fraction: float. 
        Fraction of the image to use for the registration. 
        Selection start from the center of the image. 
        Default 0.2

    """

    # Inputs of the function
    parser = argparse.ArgumentParser(description='Register the stitched images \
                                    of the reference channels')

    parser.add_argument('-path', help='path to the folder with the stitched \
                        XX.sf.hdf5 and XX_data_reg.pkl files')
    parser.add_argument('-reference_gene', help='Reference gene used for the \
                        stitching')
    
    parser.add_argument('-fraction',default=0.2, help='fraction of the picture to use for \
                        registration',type=float)

    
    args = parser.parse_args()

    # retrieve the parameters
    processing_directory = args.path
    reference_gene = args.reference_gene
    fraction = args.fraction

    # Determine the operating system running the code
    os_windows, add_slash = utils.determine_os()

    # Check training slash in the processing directory
    processing_directory=utils.check_trailing_slash(processing_directory,os_windows)



    hr.register_final_images_reg_data_only(processing_directory, gene=reference_gene,
                          sub_pic_frac=fraction, use_MPI=False,
                          apply_to_corners=True, apply_warping = False,
                          region=None, compare_in_seq=True)




if __name__ == "__main__":
    reference_registration()


from pysmFISH import utils
import argparse

def add_coords_to_experimental_metadata():

    """
    Script used to add the coords to the Experimental_metadata.yaml file.
    In order to work: 

    - The Experimental_metadata.yaml file needs to have the processing hybridization 
    and the coords at point 0 initialized: ex:  

    TilesPositions:  
    Hybridization1:  
        0:  

    - The coords file extracted from the Nikon microscope need to be in a text file
    ex. Hyb1_Coords.txt. The Coords tag is necessary and Hyb is the hybridization_key

    - The input parameters are passed via argparse

    Parameters:
    -----------
    hybridization_number: string
        Hybridization number (ex. 3) describing the hybridization to process.
    
    path: string 
        Exact path to the folder containing the text file with the coordinates.
    
    hybridization_key: string 
        Possible values 'Hyb' or 'Strip'. To add coordinates for stripping if necessary.

    """



    # Inputs of the function
    parser = argparse.ArgumentParser(description='Add microscope coords to \
                                    experimental_metadata.yaml file')
    parser.add_argument('-hybridization_number', help='hybridization to add the coords to')
    parser.add_argument('-path', help='path to the folder with the experimental_metadata.yaml file')
    parser.add_argument('-hybridization_key', default='Hyb',help='String to identify the specific \
                        coords file')
    args = parser.parse_args()

    # retrieve the parameters
    processing_directory = args.path
    hyb_nr = args.hybridization_number
    hyb_key = args.hybridization_key

    # Determine the operating system running the code
    os_windows, add_slash = utils.determine_os()

    # Check training slash in the processing directory
    processing_directory=utils.check_trailing_slash(processing_directory,os_windows)

    utils.add_coords_to_yaml(processing_directory,hyb_nr,hyb_key)



if __name__ == "__main__":
    add_coords_to_experimental_metadata()





import os
import sys
import numpy as np
import argparse
import h5py
import glob
import multiprocessing
from distributed import Client,LocalCluster
import logging
import re

# Own imports
from pysmFISH.stitching_package import stitching
from pysmFISH.stitching_package import tilejoining
from pysmFISH.stitching_package import hdf5preparation
from pysmFISH import utils



def apply_stitching():

    """
    Script to apply the registration to all the osmFISH channels. It will create
    a stitched image in an hdf5 file

    All the parameters are entered via argparse

    Parameters:
    -----------

    experiment_path: string
        Path to the folder with the hybridizations
    reference_files_path: string
        Path to the folder with the _reg_data.pkl files
    scheduler: string
        tcp address of the dask.distributed scheduler (ex. tcp://192.168.0.4:7003). 
        default = False. If False the process will run on the local computer using nCPUs-1


    """

    parser = argparse.ArgumentParser(description='Create the stitched images \
                                    after registration')

    parser.add_argument('-experiment_path', help='path to the folder with the hybridizations')
    parser.add_argument('-reference_files_path', help='path to the folder with the \
                        _reg_data.pkl files')
    parser.add_argument('-scheduler', default=False, help='dask scheduler address ex. tcp://192.168.0.4:7003')
    args = parser.parse_args()

    processing_experiment_directory = args.experiment_path
    stitched_reference_files_dir = args.reference_files_path
    # Dask scheduler address
    scheduler_address = args.scheduler

    if scheduler_address:
        # Start dask client on server or cluster
        client=Client(scheduler_address)

    else:
        # Start dask client on local machine. It will use all the availabe
        # cores -1

        # number of core to use
        ncores = multiprocessing.cpu_count()-1
        cluster = LocalCluster(n_workers=ncores)
        client=Client(cluster)


    # Determine the operating system running the code
    os_windows, add_slash = utils.determine_os()

    # Check training slash in the processing directory
    processing_experiment_directory=utils.check_trailing_slash(processing_experiment_directory,os_windows)

    stitched_reference_files_dir=utils.check_trailing_slash(stitched_reference_files_dir,os_windows)

    # Starting logger
    utils.init_file_logger(processing_experiment_directory)
    logger = logging.getLogger()


    # Collect the infos of the experiment and the processing
    # Parse the Experimental metadata file (serial)
    experiment_infos,image_properties, hybridizations_infos, \
    converted_positions, microscope_parameters =\
    utils.experimental_metadata_parser(processing_experiment_directory)

    # Parse the configuration file 
    flt_rawcnt_config = utils.filtering_raw_counting_config_parser(processing_experiment_directory)

    # Get the reference gene used
    reference_gene = flt_rawcnt_config['reference_gene']

    # Stitch the image in 2D or 3D (3D need more work/testing)
    nr_dim = flt_rawcnt_config['nr_dim']

    # Determine the hybridizations to process
    if isinstance(flt_rawcnt_config['hybs_to_stitch'],list):
        hybridizations_to_process = flt_rawcnt_config['hybs_to_stitch']
    else:
        if flt_rawcnt_config['hybs_to_stitch'] == 'All':
                hybridizations_to_process = list(hybridizations_infos.keys())
        
        else:
            raise ValueError('Error in the hybridizations to stitch')
    


    for hybridization in hybridizations_to_process:
        
        # Determine the genes to stitch in the processing hybridization
        genes_processing = list(hybridizations_infos[hybridization].keys())


        hyb_short = re.sub('Hybridization','hyb',hybridization)
        processing_hyb = experiment_infos['ExperimentName']+'_'+hyb_short
        hyb_dir = processing_experiment_directory+processing_hyb+add_slash

        # Create pointer of the hdf5 file that will store the stitched images
        # for the current hybridization

        tile_file_base_name = flt_rawcnt_config['analysis_name']+'_'+ processing_hyb
        stitching_file_name = tile_file_base_name + '.reg.sf.hdf5'

        data_name = (tile_file_base_name
                            + '_' + reference_gene
                            + '_stitching_data_reg')

        stitching_file= h5py.File(hyb_dir+stitching_file_name,'w',libver='latest')  # replace with 'a' as soon as you fix the error

        # Determine the tiles organization
        joining, tiles, nr_pixels, z_count, micData = stitching.get_place_tile_input_apply_npy(hyb_dir,stitched_reference_files_dir,data_name,image_properties,nr_dim)

        for gene in genes_processing:
        
            # Create the hdf5 file structure
            stitched_group, linear_blending, blend =  hdf5preparation.create_structures_hdf5_stitched_ref_gene_file_npy(stitching_file, joining, nr_pixels,
                                            gene, blend = 'non linear')

            # Fill the hdf5 containing the stitched image with empty data and
            # create the blending mask
            stitched_group['final_image'][:]= np.zeros(joining['final_image_shape'],dtype=np.uint16)
            if blend is not None:
                # make mask
                stitched_group['blending_mask'][:] = np.zeros(joining['final_image_shape'][-2:],dtype=np.uint16)
                tilejoining.make_mask(joining, nr_pixels, stitched_group['blending_mask'])

            filtered_img_gene_dirs_path = hyb_dir+flt_rawcnt_config['analysis_name']+'_'+processing_hyb +'_filtered_npy'+add_slash
            filtered_img_gene_dirs = glob.glob(filtered_img_gene_dirs_path+'*')

            # Create the subdirectory used to save the blended tiles
            suffix = 'blended_tiles'
            blended_tiles_directory = utils.create_single_directory(hyb_dir,gene, hybridization,processing_hyb,suffix,add_slash,
                                            analysis_name=flt_rawcnt_config['analysis_name'])

            # Get the directory with the filtered npy images of the reference_gene to use for stitching
            stitching_files_dir = [npy_dir for npy_dir in filtered_img_gene_dirs if gene in npy_dir][0]
            stitching_files_dir= stitching_files_dir+add_slash

            # Create the tmp directory where to save the masks
            suffix = 'masks'
            masked_tiles_directory = utils.create_single_directory(hyb_dir,gene,hybridization,processing_hyb,suffix,add_slash,
                                            analysis_name=flt_rawcnt_config['analysis_name'])

            # Create and save the mask files
            for corn_value,corner_coords in joining['corner_list']:
                if not(np.isnan(corner_coords[0])):
                    cur_mask = stitched_group['blending_mask'][int(corner_coords[0]):int(corner_coords[0]) + int(nr_pixels),
                                        int(corner_coords[1]):int(corner_coords[1]) + int(nr_pixels)]

                    fname = masked_tiles_directory + flt_rawcnt_config['analysis_name'] +'_'+processing_hyb+'_'+gene+'_masks_joining_pos_'+str(corn_value)
                    np.save(fname,cur_mask)


            # Blend all the tiles and save them in a directory
            futures_processes = client.map(tilejoining.generate_blended_tile_npy,joining['corner_list'],
                                        stitching_files_dir = stitching_files_dir,
                                        blended_tiles_directory = blended_tiles_directory,
                                        masked_tiles_directory = masked_tiles_directory,
                                        analysis_name = flt_rawcnt_config['analysis_name'],
                                        processing_hyb = processing_hyb,reference_gene = gene,
                                        micData = micData,tiles = tiles,nr_pixels=nr_pixels,
                                        linear_blending=linear_blending)



            _ = client.gather(futures_processes)


            # Write the stitched image
            tilejoining.make_final_image_npy(joining, stitching_file, blended_tiles_directory, tiles,gene, nr_pixels)
            stitching_file.flush()

            # Remove directories with blended tiles and masks
            shutil.rmtree(blended_tiles_directory)
            shutil.rmtree(masked_tiles_directory)


        stitching_file.close()

    
    client.close()


if __name__ == "__main__":
    apply_stitching()